/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/encoding/insertutil.c
 * @brief Break file that is inserted into blocks and encrypts
 *        them according to the CHK-triple-hash-tree scheme (ESED II).
 * @see http://www.ovmj.org/GNUnet/encoding.php3
 * @author Krista Bennett
 * @author Christian Grothoff
 **/

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/storage.h"
#include "util/tcp_return.h"
#include "util/xmalloc.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/util/contentencoding.h"

/**
 * Ask gnunetd to append a filename.
 * 
 * @param sock connection to gnunetd
 * @param filename the name to add to fileindex.c
 * @return the index, -1 on error
 **/
static int askAppendFilename(GNUNET_TCP_SOCKET * sock,
			     char * filename) {
  AFS_CS_INDEX_FILE * request;
  int result;
  
  filename 
    = expandFileName(filename);   
  request = MALLOC(sizeof(AFS_CS_INDEX_FILE));
  request->header.size 
    = htons(sizeof(AFS_CS_INDEX_FILE));
  request->header.tcpType 
    = htons(AFS_CS_PROTO_INDEX_FILE);
  memset(&request->filename[0], 
	 0,
	 CS_FILE_LIST_FILENAME);
  memcpy(&request->filename[0],
	 filename,
	 strlen(filename));
  FREE(filename);
  if (SYSERR == writeToSocket(sock,
			      &request->header)) {
    FREE(request);
    LOG(LOG_WARNING, 
	"WARNING: could not send data to gnunetd. "\
	"Is gnunetd running?\n");
    return -1;
  } 
  FREE(request);
  if (SYSERR == readTCPResult(sock,
			      &result)) {
    LOG(LOG_WARNING, 
	"WARNING: could not receive data from gnunetd. Is gnunetd running?\n");
    result = -1;
  }
  return result;
}
  
/**
 * Creates root node for the tree and writes the top-level tree node.
 *
 * @param sock connection to gnunetd
 * @param rn the RootNode to insert
 * @param keyword the keyword under which the rn is inserted
 * @param contentPriority priority of the inserted content
 **/
static int insertRootWithKeyword(GNUNET_TCP_SOCKET * sock,
				 RootNode * rn,
				 char * keyword,
				 int contentPriority) {
  HashCode160 hc;
  AFS_CS_INSERT_3HASH * msg;
  int res;
  
  hash(keyword, 
       strlen(keyword), 
       &hc);
  msg = MALLOC(sizeof(AFS_CS_INSERT_3HASH));
  if (SYSERR == encryptContent((CONTENT_Block*)rn,
			       &hc,
			       &msg->content)) 
    errexit("Encryption failed.\n"); 

  hash(&hc,
       sizeof(HashCode160),
       &msg->doubleHash);
  msg->importance 
    = htonl(contentPriority);
  msg->header.tcpType 
    = htons(AFS_CS_PROTO_INSERT_3HASH);
  msg->header.size 
    = htons(sizeof(AFS_CS_INSERT_3HASH));
  if (SYSERR == writeToSocket(sock,
			      &msg->header)) {
    LOG(LOG_WARNING, 
	"WARNING: could not send data to gnunetd. "\
	"Is gnunetd running?\n");
    FREE(msg);
    return SYSERR;
  }
  FREE(msg);
  if (SYSERR == readTCPResult(sock,
			      &res)) {
    LOG(LOG_WARNING, 
	"WARNING: server did not send confirmation of insertion\n");
    return SYSERR;
  } else
    if (res == SYSERR)
      LOG(LOG_WARNING, 
	  "WARNING: server could not perform insertion\n");
  return res;
}

/**
 * Insert (or index) a file under the given name into the local GNUnet
 * node.
 *
 * @param sock connection to gnunetd
 * @param filename the name of the file to insert
 * @param model the insert model used to
 *        update status information; points to NULL if
 *        no status updates shall be given, otherwise 
 *        to a method that takes two size_t arguments
 *        (retrieved so far, total).
 * @param model_data pointer that is passed to the model method
 * @return NULL on error, otherwise the top block
 **/
Block * insertFile(GNUNET_TCP_SOCKET * sock,
		   char * filename, 
		   ProgressModel model,
		   void * model_data) {
  NodeContext nc;
  size_t filesize;
  Block * top;

  filename = expandFileName(filename);
  filesize = (size_t) getFileSize(filename);
  nc.pmodel = model;
  nc.data = model_data;
  memset(&nc.stats, 0, sizeof(ProgressStats));
  nc.stats.filesize = filesize;
  nc.priority = getConfigurationInt("GNUNET-INSERT",
				    "CONTENT-PRIORITY");
  if (nc.priority == 0)
    nc.priority = LOCAL_INDEXED_CONTENT_PRIO;
  if (YES == testConfigurationString("GNUNET-INSERT",
				     "INDEX-CONTENT",
				     "YES")) {
    nc.index = askAppendFilename(sock, filename);
    if (nc.index == 0)
      LOG(LOG_WARNING,
	  "WARNING: adding to index list failed, trying insertion!\n");
  } else 
    nc.index = 0; /* 0: no indexing */
  if (SYSERR == createIOContext(&nc.ioc,
				filesize,
				filename,
				YES)) {    
    FREE(filename);
    return NULL;
  }
  if (filesize <= sizeof(CONTENT_Block))
    top = createTopDBlock(filesize);
  else
    top = createTopIBlock(filesize);
  if (SYSERR == top->insert(top, &nc, sock)) {
    top->done(top, NULL);
    freeIOC(&nc.ioc, NO);
    FREE(filename);
    return NULL;
  }
  freeIOC(&nc.ioc, NO);

  FREE(filename);
  return top;
}

/**
 * Insert a root-block into GNUnet. 
 *
 * @param sock connection to gnunetd
 * @param top the top block of the file
 * @param description description to use
 * @param filenameRoot filename to use
 * @param mimetype mimetype to use
 * @param keywords the keywords that shall be used to retrieve the file
 * @param num_keys the number of keywords to be associated with the file
 * @return OK on success, SYSERR on error
 **/
int insertRoot(GNUNET_TCP_SOCKET * sock,
	       Block * top,
	       char * description,
     	       char * filenameRoot,
	       char * mimetype,
	       unsigned int num_keys,
	       char ** keywords) {
  int i;
  int priority;
  RootNode * rn;
  int res;

  priority = getConfigurationInt("GNUNET-INSERT",
				 "CONTENT-PRIORITY");
  rn = MALLOC(sizeof(RootNode));
  rn->header.major_formatVersion 
    = htons(ROOT_MAJOR_VERSION);
  rn->header.minor_formatVersion 
    = htons(ROOT_MINOR_VERSION);
  rn->header.fileIdentifier.file_length
    = htonl(top->filesize);
  rn->header.fileIdentifier.crc 
    = htonl(crc32N(top->data, top->len));
  memcpy(&rn->header.fileIdentifier.chk,
	 &top->chk,
	 sizeof(CHK_Hashes));
  if (strlen(description) >= MAX_DESC_LEN)
    description[MAX_DESC_LEN-1] = 0;
  memcpy(&rn->header.description[0],
	 description,
	 strlen(description)+1);
  if (strlen(filenameRoot) >= MAX_FILENAME_LEN)
    filenameRoot[MAX_FILENAME_LEN-1] = 0;
  memcpy(&rn->header.filename[0],
	 filenameRoot,
	 strlen(filenameRoot)+1);
  if (strlen(mimetype) >= MAX_MIMETYPE_LEN)
    description[MAX_MIMETYPE_LEN-1] = 0;
  memcpy(&rn->header.mimetype[0],
	 mimetype,
	 strlen(mimetype)+1);
  res = OK;
  for (i=0;i<num_keys;i++) 
    if (SYSERR == insertRootWithKeyword(sock,
					rn, 
					keywords[i], 
					priority))
      res = SYSERR;
  FREE(rn);
  return res;
}

/* end of insertutil.c */
