/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/xmalloc.c
 * @brief wrapper around malloc/free 
 * @author Christian Grothoff
 **/ 

#include "config.h"
#include "util/logging.h"
#include "util/semaphore.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define DEBUG_MALLOC 0 

#if DEBUG_MALLOC
static Mutex lock;
#endif

void initXmalloc() {
#if DEBUG_MALLOC
  MUTEX_CREATE(&lock);
#endif
}

void doneXmalloc() {
#if DEBUG_MALLOC
  MUTEX_DESTROY(&lock);
#endif
}

/**
 * Allocate memory. Checks the return value, aborts if no more
 * memory is available.
 *
 * @param size how many bytes of memory to allocate
 * @param filename where in the code was the call to GROW
 * @param linenumber where in the code was the call to GROW
 * @return pointer to size bytes of memory
 **/
void * xmalloc_(int size, 
		const char * filename,
		const int linenumber) {
  void * result;

  result = malloc(size);
#if DEBUG_MALLOC 
  MUTEX_LOCK(&lock);
  printf("%x malloc %s:%d (%d bytes)\n", 
	 (int)result, 
	 filename,
	 linenumber,
	 size);
  fflush(stdout);
  MUTEX_UNLOCK(&lock);
#endif
  if (result == NULL) {
    errexit("ERROR: out of memory (%d at %s:%d)!\n",
	    size, 
	    filename,
	    linenumber);
  }
  memset(result, 0, size); /* client code should not rely on this, though... */
  return result;
}

/**
 * Free memory. Merely a wrapper for the case that we
 * want to keep track of allocations.
 *
 * @param ptr the pointer to free
 * @param filename where in the code was the call to GROW
 * @param linenumber where in the code was the call to GROW
 **/
void xfree_(void * ptr, 
	    const char * filename,
	    const int linenumber) {
  if (ptr == NULL)
    errexit("FREE called with NULL ptr at %s:%d\n", 
	    filename,
	    linenumber);
#if DEBUG_MALLOC 
  MUTEX_LOCK(&lock);
  printf("%x free %s:%d\n", 
	 (int) ptr, 
	 filename,
	 linenumber);
  fflush(stdout);
  MUTEX_UNLOCK(&lock);
#endif
  free(ptr);
}

/**
 * Dup a string (same semantics as strdup).
 *
 * @param str the string to dup
 * @param filename where in the code was the call to GROW
 * @param linenumber where in the code was the call to GROW
 * @return strdup(str)
 **/
char * xstrdup_(const char * str,
		const char * filename,
		const int linenumber) {
  char * res;

  res = strdup(str);
#if DEBUG_MALLOC 
  MUTEX_LOCK(&lock);
  printf("%x malloc %s:%d (%d bytes, strdup of %s)\n", 
	 (int) res, 
	 filename,
	 linenumber,
	 strlen(str)+1,
	 str);
  fflush(stdout);
  MUTEX_UNLOCK(&lock);
#endif
  return res;
}

/**
 * Grow an array.  Grows old by (*oldCount-newCount)*elementSize bytes
 * and sets *oldCount to newCount.
 *
 * @param old address of the pointer to the array
 *        *old may be NULL
 * @param elementSize the size of the elements of the array
 * @param oldCount address of the number of elements in the *old array
 * @param newCount number of elements in the new array, may be 0
 * @param filename where in the code was the call to GROW
 * @param linenumber where in the code was the call to GROW
 **/
void xgrow_(void ** old,
	    int elementSize,
	    int * oldCount,
	    int newCount,
	    char * filename,
	    const int linenumber) {
  void * tmp;
  int size;

  size = newCount * elementSize;
  if (size == 0)
    tmp = NULL;
  else {
    tmp = malloc(size);  
#if DEBUG_MALLOC 
    MUTEX_LOCK(&lock);
    printf("%x malloc %s:%d (grow to %d)\n", 
	   (int)tmp, 
	   filename,
	   linenumber,
	   size);
    fflush(stdout);
    MUTEX_UNLOCK(&lock);
#endif
    if (tmp == NULL) {
      errexit("ERROR: out of memory (%d at %s:%d)!\n",
	      size, 
	      filename,
	      linenumber);
    }
    memset(tmp, 0, size); /* client code should not rely on this, though... */
    if (*oldCount > newCount)
      *oldCount = newCount; /* shrink is also allowed! */
    memcpy(tmp, 
	   *old, 
	   elementSize * (*oldCount));
  }

  if (*old != NULL) {
#if DEBUG_MALLOC 
    MUTEX_LOCK(&lock);
    printf("%x free %s:%d (grow)\n", 
	   (int) *old, 
	   filename,
	   linenumber);
    fflush(stdout);
    MUTEX_UNLOCK(&lock);
#endif
    free(*old);
  }
  *old = tmp;
  *oldCount = newCount;
}

/* end of xmalloc.c */
